
// SPDX-License-Identifier: GPL-2.0-or-later

#include "compat/compat.h"
#include "oregano_updater_altera.h"

#include <linux/debugfs.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/io.h>
#include <linux/jiffies.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>

#ifdef CONFIG_OF
#include <linux/of.h>
#include <linux/of_device.h>
#endif


#ifndef DRIVER_NAME
#define DRIVER_NAME	"oregano-updater-altera"
#endif

enum oregano_update_altera_reg_status {
	OREGANO_UPDATE_ALTERA_STATUS_ERROR = BIT(0),
	OREGANO_UPDATE_ALTERA_STATUS_ERROR_CRC = BIT(1),
	OREGANO_UPDATE_ALTERA_STATUS_ERROR_WATCHDOG = BIT(5),
	OREGANO_UPDATE_ALTERA_STATUS_FACTORY_LOADED = BIT(6),
};



struct oregano_updater_altera {
	struct device *dev;
	void __iomem *iobase; // memory base

	enum syn1588_ioctl_update_status current_operation;
	size_t addr_position; // current addr postion of the read / write process

	struct tasklet_struct tasklet;
	u32 *data; // vmalloc allocated memory
	size_t data_size; // in bytes
};

enum oregano_update_altera_regs {
	OREGANO_UPDATE_ALTERA_CTRL = 0x0u, // RW
	OREGANO_UPDATE_ALTERA_DATA = 0x4u, // RW
	OREGANO_UPDATE_ALTERA_STATUS = 0x8u, // R
};
enum oregano_update_altera_reg_ctrl {
	OREGANO_UPDATE_ALTERA_CTRL_READ_CFG0 = BIT(0),
	OREGANO_UPDATE_ALTERA_CTRL_WRITE_CFG0 = BIT(1),
	OREGANO_UPDATE_ALTERA_CTRL_ERASE_CFG0 = BIT(2),

	OREGANO_UPDATE_ALTERA_CTRL_INIT = BIT(16),
	OREGANO_UPDATE_ALTERA_CTRL_BUSY = BIT(17),
};

static const size_t SPI_MEMORY_SIZE = 0x500000;


static struct platform_driver oregano_updater_altera_driver;
struct oregano_updater_altera *dev_get_updater_altera(struct device *dev) {
	if(IS_ERR_OR_NULL(dev)) {

		return NULL;
	}
	if(dev->driver != &oregano_updater_altera_driver.driver){
		return NULL;
	}
	return dev_get_drvdata(dev);
}
EXPORT_SYMBOL_GPL(dev_get_updater_altera);

/*----------------------------------------------------------------------*/

static u32 oregano_updater_altera_get_status(struct oregano_updater_altera *oregano_updater) {
	return  ioread32(oregano_updater->iobase + OREGANO_UPDATE_ALTERA_STATUS);
}

static int oregano_updater_altera_wait_busy(struct oregano_updater_altera *oregano_updater, unsigned long timeout_in_ms) {
	unsigned long j = jiffies + timeout_in_ms;
	while( true ) {
		// wait until busy flag is gone
		u32 status = ioread32(oregano_updater->iobase + OREGANO_UPDATE_ALTERA_CTRL);
		if((status & OREGANO_UPDATE_ALTERA_CTRL_BUSY) == 0 ) {
			break;
		}
		if(timeout_in_ms > msecs_to_jiffies(1000)) {
			msleep(1000);
		}
		if(j < jiffies) {
			return -ETIMEDOUT;
		}
	}
	return 0;
}


static int oregano_updater_altera_start_write_firmware(struct oregano_updater_altera *oregano_updater) {
	struct device *dev = oregano_updater->dev;
	int timeout = msecs_to_jiffies(20000);
	int result = 0;

	dev_info(dev, "flash new bitstream over altera updater interface\n");

	iowrite32(OREGANO_UPDATE_ALTERA_CTRL_INIT, oregano_updater->iobase + OREGANO_UPDATE_ALTERA_CTRL);

	result = oregano_updater_altera_wait_busy(oregano_updater, timeout);
	if(result) {
		dev_err(dev, "timeout waiting after init\n");
		return result;
	}

	return 0;
}

static int oregano_updater_altera_process_write_firmware(struct oregano_updater_altera *oregano_updater, const u32 *data, size_t word_length) {
	size_t i;
	struct device *dev = oregano_updater->dev;
	int timeout = msecs_to_jiffies(100);
	int result = 0;
	size_t current_bitstream_size = oregano_updater->addr_position + word_length;

	if(current_bitstream_size > SPI_MEMORY_SIZE) {
		dev_err(dev, "firmwarefile is to big max %zu\n", current_bitstream_size);
	}


	dev_info(dev, "update chunk size: %dB\n", (int)word_length*4);

	for(i = 0; i < word_length; i++) {
		iowrite32(data[i], oregano_updater->iobase + OREGANO_UPDATE_ALTERA_DATA);

		if(oregano_updater->addr_position == 0) {
			iowrite32(OREGANO_UPDATE_ALTERA_CTRL_WRITE_CFG0, oregano_updater->iobase + OREGANO_UPDATE_ALTERA_CTRL);
			msleep(1); // delay because hardware has sync issues by reading the busy flag to fast
			/*
			 * Max timeout is currently from the dual nic v1.0
			 * Model        | max time sector | sectors   | total
			 * rev2.1       | 3s              | 68        | 204s
			 * rev2.3       | 1s              | 108 or 68 | 108s
			 * dualnic rev1 | 3s              | 108       | 324s
			 */
			dev_info(dev, "start erase could take up to 324s\n");
			result = oregano_updater_altera_wait_busy(oregano_updater, 108 * 3 * 1000);
			if(result) {
				dev_err(dev, "error while update, timeout erease memory\n");
				return result;
			}
		}

		udelay(20); // delay because hardware has sync issues by reading the busy flag to fast
		result = oregano_updater_altera_wait_busy(oregano_updater, timeout);
		if(result) {
			dev_err(dev, "error while update, timeout writing data at addr: %08x\n", (int)i*4);
			return result;
		}
		oregano_updater->addr_position++;
	}
	return 0;
}


static int oregano_updater_altera_start_validate_firmware(struct oregano_updater_altera *oregano_updater) {
	struct device *dev = oregano_updater->dev;
	int timeout = msecs_to_jiffies(100);
	int result = 0;

	oregano_updater->addr_position = 0;

	iowrite32(OREGANO_UPDATE_ALTERA_CTRL_INIT, oregano_updater->iobase + OREGANO_UPDATE_ALTERA_CTRL);

	result = oregano_updater_altera_wait_busy(oregano_updater, timeout);
	if(result) {
		dev_err(dev, "timeout waiting after init\n");
		return result;
	}

	iowrite32(OREGANO_UPDATE_ALTERA_CTRL_READ_CFG0, oregano_updater->iobase + OREGANO_UPDATE_ALTERA_CTRL);

	return 0;
}

static int oregano_updater_altera_process_validate_firmware(struct oregano_updater_altera *oregano_updater, const u32 *data, size_t data_word_length) {
	size_t i;
	struct device *dev = oregano_updater->dev;
	int timeout = msecs_to_jiffies(100);
	int result = 0;
	dev_info(dev, "read bitstream over altera updater interface\n");


	for(i = 0; i < data_word_length; i++) {
		u32 data_read;
		result = oregano_updater_altera_wait_busy(oregano_updater, timeout);
		if(result) {
			dev_err(dev, "error while update, timeout reading data at addr: %08x\n", (int)i*4);
			return result;
		}
		data_read = ioread32(oregano_updater->iobase + OREGANO_UPDATE_ALTERA_DATA);
		if(data[i] != data_read) {
			return -EIO;
		}
		oregano_updater->addr_position += sizeof(data_read);
	}
	return 0;
}


int oregano_updater_altera_update_firmware(struct oregano_updater_altera *oregano_updater, size_t data_size, u32 *data)
{
	struct device *dev = oregano_updater->dev;

	switch(oregano_updater->current_operation) {
	case SYN1588_IOCTL_UPDATE_STATUS_IDLE:
	case SYN1588_IOCTL_UPDATE_STATUS_SUCCESS:
	case SYN1588_IOCTL_UPDATE_STATUS_FAILED:
		break;
	default:
		dev_err(dev, "could not start firmware update, as other update is in progress\n");
		return -EBUSY;
	}
	oregano_updater->data = data;
	oregano_updater->data_size = data_size;

	oregano_updater->addr_position = 0;
	oregano_updater->current_operation = SYN1588_IOCTL_UPDATE_STATUS_STARTING;

	tasklet_schedule(&oregano_updater->tasklet);

	return 0;
}
EXPORT_SYMBOL_GPL(oregano_updater_altera_update_firmware);

struct syn1588_ioctl_update_status_reponse oregano_updater_altera_update_status(struct oregano_updater_altera *oregano_updater)
{
	struct syn1588_ioctl_update_status_reponse response = {
		oregano_updater->current_operation,
		oregano_updater->addr_position * 100 /  oregano_updater->data_size,
	};

	return response;
}
EXPORT_SYMBOL_GPL(oregano_updater_altera_update_status);

static void oregano_updater_altera_check_status(struct oregano_updater_altera *oregano_updater) {
	u32 status_reg = oregano_updater_altera_get_status(oregano_updater);

	if(status_reg != 0) {
		dev_warn(oregano_updater->dev, "User bitstream not loaded, status register: %04X\n", status_reg);
	}
}

static void oregano_updater_altera_task(unsigned long arg) {

	struct oregano_updater_altera *oregano_updater = (void *)arg;
	struct device *dev = oregano_updater->dev;
	const u32 *data = oregano_updater->data;
	const size_t data_size = oregano_updater->data_size;
	const size_t chunk_size = 1024 * 8; // 8kB
	size_t bitstream_pos = 0;
	int result;

	oregano_updater->data = NULL;

	if(oregano_updater->current_operation != SYN1588_IOCTL_UPDATE_STATUS_STARTING) {
		dev_err(dev, "invalid state abort update\n");
		return;
	}
	oregano_updater->current_operation = SYN1588_IOCTL_UPDATE_STATUS_PROGRESS;

	result = oregano_updater_altera_start_write_firmware(oregano_updater);
	if(result) {
		dev_err(dev, "could not start with write firmware\n");
		vfree(data);
		oregano_updater->current_operation = SYN1588_IOCTL_UPDATE_STATUS_FAILED;
		return;
	}
	for(bitstream_pos = 0; bitstream_pos < data_size; bitstream_pos+=chunk_size) {
		size_t data_chunk_size = min(chunk_size, data_size - bitstream_pos);
		size_t data_chunk_word_size = DIV_ROUND_UP(data_chunk_size, 4);
		oregano_updater->addr_position = bitstream_pos;
		dev_info(dev, "update progress: %d\n", (int)(bitstream_pos*100 / data_size) );
		msleep(10); // allow the kernel to schedule this thread

		result = oregano_updater_altera_process_write_firmware(
			oregano_updater,
			data + bitstream_pos/sizeof(data[0]),
			data_chunk_word_size);

		if(result) {
			dev_err(dev, "could not start with write firmware\n");
			vfree(data);
			oregano_updater->current_operation = SYN1588_IOCTL_UPDATE_STATUS_FAILED;
			return;
		}
	}

	oregano_updater->current_operation = SYN1588_IOCTL_UPDATE_STATUS_PROGRESS_VERIFY;

	result = oregano_updater_altera_start_validate_firmware(oregano_updater);
	if(result) {
		dev_err(dev, "could not start with validating update\n");
		vfree(data);
		oregano_updater->current_operation = SYN1588_IOCTL_UPDATE_STATUS_FAILED;
		return;
	}
	for(bitstream_pos = 0; bitstream_pos < data_size; bitstream_pos+=chunk_size) {
		size_t data_chunk_size = min(chunk_size, data_size - bitstream_pos);
		size_t data_chunk_word_size = DIV_ROUND_UP(data_chunk_size, 4);

		dev_info(dev, "read progress: %d %%\n", (int)(bitstream_pos*100 / data_size) );
		msleep(10); // allow the kernel to schedule this thread

		result = oregano_updater_altera_process_validate_firmware(oregano_updater, data + bitstream_pos/4, data_chunk_word_size);
		if(result) {
			dev_err(dev, "error in validate firmware\n");
			vfree(data);
			oregano_updater->current_operation = SYN1588_IOCTL_UPDATE_STATUS_FAILED;
			return;
		}
	}

	vfree(data);
	oregano_updater->current_operation = SYN1588_IOCTL_UPDATE_STATUS_SUCCESS;
}


static int oregano_updater_altera_probe_pdata(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct oregano_updater_altera_platform_data *pdata = dev_get_platdata(dev);
	(void)pdata;

	return 0;
}


static int oregano_updater_altera_remove(struct platform_device *pdev) {
	struct oregano_updater_altera * oregano_updater = dev_get_drvdata(&pdev->dev);
	tasklet_kill(&oregano_updater->tasklet);
	return 0;
}

static int oregano_updater_altera_probe(struct platform_device *pdev)
{
	int				status = 0;
	struct oregano_updater_altera	*oregano_updater;
	struct device			*dev = &pdev->dev;
	struct resource *res;

	dev_info(dev, "probe oregano updater altera driver\n");

	if (!dev_of_node(&pdev->dev))
		status = oregano_updater_altera_probe_pdata(pdev);

	if (status)
		return status;


	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if(res == NULL) {
		dev_err(dev, "could not request resources\n");
		return -ENODEV;
	}
	oregano_updater = kzalloc(sizeof(struct oregano_updater_altera),GFP_KERNEL);
	if(IS_ERR_OR_NULL(oregano_updater)) {
		return -ENOMEM;
	}
	platform_set_drvdata(pdev, oregano_updater);
	oregano_updater->dev = &pdev->dev;

	oregano_updater->iobase = devm_ioremap_resource(dev, pdev->resource);
	if(IS_ERR_OR_NULL(oregano_updater->iobase)) {
		dev_err(dev, "could not ioremap resource\n");
		return PTR_ERR(oregano_updater->iobase);
	}

	tasklet_init(&oregano_updater->tasklet, oregano_updater_altera_task, (unsigned long)oregano_updater);

	oregano_updater_altera_check_status(oregano_updater);


	return 0;
}

MODULE_ALIAS("platform:" DRIVER_NAME);

static struct platform_driver oregano_updater_altera_driver = {
	.driver = {
		.name	= DRIVER_NAME,
	},
	.probe		= oregano_updater_altera_probe,
	.remove		= oregano_updater_altera_remove,
};
module_platform_driver(oregano_updater_altera_driver);

MODULE_DESCRIPTION("update driver for oregano altera devices");
MODULE_AUTHOR("Norbert Summer");
MODULE_LICENSE("GPL");
