
// SPDX-License-Identifier: GPL-2.0-or-later
#include "altera_pci_bridge.h"

#include <linux/device.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/slab.h>

#include "compat/compat.h"


#ifndef DRIVER_NAME
#define DRIVER_NAME	"altera-pci-bridge"
#endif


enum syn1588_pci_bridge {
	DMA_TRANSLATE0_L = 0x1000,
	DMA_TRANSLATE0_H = 0x1004,
};


struct altera_pci_bridge_controller {
	void __iomem *iobase; // memory base
	u32 address_mask; // mask for host part
};

static struct platform_driver altera_pci_bridge_driver;


void altera_pci_bridge_set_address_translation(struct device*dev, dma_addr_t addr) {
	struct altera_pci_bridge_controller *pci_bridge = dev_get_drvdata(dev);
	if(dev->driver != &altera_pci_bridge_driver.driver) {
		pr_err("invalid driver type for this device\n");
		return;
	}
	if(IS_ERR_OR_NULL(pci_bridge)) {
		pr_err("pci bridge is not initialized\n");
		return;
	}

	iowrite32((u32)addr & pci_bridge->address_mask, pci_bridge->iobase + DMA_TRANSLATE0_L);
	iowrite32(0x0, pci_bridge->iobase + DMA_TRANSLATE0_H);

}
EXPORT_SYMBOL_GPL(altera_pci_bridge_set_address_translation);

u32 altera_pci_bridge_get_address_mask(struct device*dev) {
	struct altera_pci_bridge_controller *pci_bridge = dev_get_drvdata(dev);
	if(dev->driver != &altera_pci_bridge_driver.driver) {
		pr_err("invalid driver type for this device\n");
		return 0x0;
	}
	if(IS_ERR_OR_NULL(pci_bridge)) {
		pr_err("pci bridge is not initialized\n");
		return 0x0;
	}

	return pci_bridge->address_mask;

}
EXPORT_SYMBOL_GPL(altera_pci_bridge_get_address_mask);

static int altera_pci_bridge_remove(struct platform_device *pdev) {
	struct altera_pci_bridge_controller *pci_bridge = platform_get_drvdata(pdev);

	if(IS_ERR_OR_NULL(pci_bridge)) {
		return 0;
	}

	kfree(pci_bridge);
	platform_set_drvdata(pdev, NULL);
	return 0;
}

static int altera_pci_bridge_probe(struct platform_device *pdev)
{
	struct altera_pci_bridge_controller *pci_bridge;
	struct device *dev = &pdev->dev;
	struct resource *res;
	u32 passthru_mask = ~0;

	dev_info(dev, "probe dma translation driver\n");


	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if(res == NULL) {
		dev_err(dev, "could not request resources\n");
		return -ENODEV;
	}
	pci_bridge = kzalloc(sizeof(struct altera_pci_bridge_controller), GFP_KERNEL);
	platform_set_drvdata(pdev, pci_bridge);

	pci_bridge->iobase = devm_ioremap_resource(dev, pdev->resource);
	if(IS_ERR_OR_NULL(pci_bridge->iobase)) {
		dev_err(dev, "could not ioremap resource\n");
		altera_pci_bridge_remove(pdev);
		return PTR_ERR(pci_bridge->iobase);
	}


	iowrite32(passthru_mask, pci_bridge->iobase + DMA_TRANSLATE0_L);
	passthru_mask = ioread32(pci_bridge->iobase + DMA_TRANSLATE0_L);
	passthru_mask &= 0xFFFFFFFC; // filter status bits
	dev_dbg(dev, "DMA passthru mask: %08x\n", passthru_mask);
	pci_bridge->address_mask = passthru_mask;



	dev_info(dev, "%s(): pci bridge unit init OK\n", __func__);
	return 0;
}

MODULE_ALIAS("platform:" DRIVER_NAME);

static struct platform_driver altera_pci_bridge_driver = {
	.driver = {
		.name	= DRIVER_NAME,
	},
	.probe		= altera_pci_bridge_probe,
	.remove		= altera_pci_bridge_remove,
};
module_platform_driver(altera_pci_bridge_driver);

MODULE_DESCRIPTION("altera pci bridge driver for oregano");
MODULE_AUTHOR("Norbert Summer");
MODULE_LICENSE("GPL");
