
#include "clock_m.h"
#include "syn1588nic_local.h"
#include "clock_m_phc.h"
#include "SyncDriver.h"
#include "oregano_tse_mac.h"
#include "compat/compat.h"
#include "version.h"

#include <linux/i2c.h>
#ifndef RHEL_BACKPORT_IDR2
#include <linux/idr.h>
#else
#include <linux/idr2.h>
#endif
#include <linux/math64.h>
#include <linux/err.h>
#include <linux/math64.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/printk.h>
#include <linux/platform_device.h>

#if defined(CONFIG_OF)
#include <linux/of.h>
#include <linux/of_device.h>
#endif


#define DRV_NAME "oregano-clock"

bool syn1588_disabled_output = false;
module_param(syn1588_disabled_output, bool, 0);
MODULE_PARM_DESC(syn1588_disabled_output, "all outputs are disabled by default");

/*
 * Clock_M
 */


#define DEV_NAME_TPL "clockm%d" ///< template for generating device file

struct clock_m_device *dev_get_clock_m(struct device *dev) {
	if(IS_ERR_OR_NULL(dev)) {
		return NULL;
	}
	//TODO check driver
	return dev_get_drvdata(dev);
}
EXPORT_SYMBOL_GPL(dev_get_clock_m);

resource_size_t clock_m_memory_size(struct clock_m_device *clock_m) {
	return resource_size(clock_m->resource);
}

// Read some data from a port
static u32 clock_m_readl(struct clock_m_device *dev, u32 offset)
{
	return ioread32(dev->iobase + offset);
}
// Write some data to a port
static void clock_m_writel(struct clock_m_device *dev, u32 val, u32 offset)
{
	iowrite32(val, dev->iobase + offset);
}


static void syn1588_BHH_CLOCK_M(unsigned long);



/**
 * Function to read from a CLOCK_M register
 * \param clock_m_dev pointer to clock_m_device structure
 * \param address register address to read from
 * \param data pointer to buffer where read data is stored
 * \return 0 on success, negative values if an error occured
 * \note Don't read the IRSRC register via this exported function! (only via
 * ioctl!)
 *			 register a callback for interrupt notification instead
 * \ingroup drv_ifc
 */
int clock_m_getreg(struct clock_m_device *clock_m_dev, unsigned long address, u32 *data)
{

	// set the lower bits to zero, since only reading from that addresses makes
	// sense
	address &= ~((u32)0b11);

	switch (address) {
	case IRSRC:
		spin_lock_bh(&clock_m_dev->reg_lock);
		*data = clock_m_dev->chr_irsrc;
		spin_unlock_bh(&clock_m_dev->reg_lock);
		pr_debug("read from Clock IRSRC data: 0x%lX\n",
			 (unsigned long)*data);
		break;
	default:
		if (address >= 0x0 && address < clock_m_memory_size(clock_m_dev)) {
			*data = clock_m_readl(clock_m_dev, address);
			pr_debug(
				 "read from Clock address 0x%lX data: 0x%lX\n",
				 address, (unsigned long)*data);
		} else {
			// read invalid virtual register address
			pr_warn(
				 "unexpected read access to virtual register: 0x%lX\n",
				 address);
			*data = 0;
			return -EINVAL;
		}
		break;
	}
	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_getreg);


/**
 * Function to read from a CLOCK_M register
 * \param clock_m_dev pointer to clock_m_device structure
 * \param address register address to read from
 * \param data to write to register specified by address
 * \return 0 on success, negative values if an error occured
 * \note Don't write to the IRSRC register via this exported function! (only via
 * ioctl!)
 * \ingroup drv_ifc
 */
int clock_m_setreg(struct clock_m_device *clock_m_dev, unsigned long address, u32 data)
{
	struct device *dev = clock_m_dev->dev;
	// set the lower bits to zero, since only reading from that addresses makes
	// sense
	address &= ~((u32)0b11);

	dev_dbg(dev, "clock_m_setreg: clock_m: 0x%p, addr: %08lx, data: %08x\n", clock_m_dev, address, (unsigned int)data);

	switch (address) {
	case IRSRC:
		spin_lock_bh(&clock_m_dev->reg_lock);
		clock_m_dev->chr_irsrc = data;
		spin_unlock_bh(&clock_m_dev->reg_lock);
		pr_debug("write to CLOCK_M IRSRC data: 0x%lX\n",
			 (unsigned long)data);
		return 0;
#ifdef USE_MBG_HPS
	case VR_TSU_PHY_INFO:
		mbg_set_tsu_phy_info(data);
		break;
#endif
	default:
		if (address >= 0x0 && address < clock_m_memory_size(clock_m_dev)) {
			pr_debug("write to CLOCK_M address 0x%lX data: 0x%lX\n",
				 address, (unsigned long)data);

			clock_m_writel(clock_m_dev, data, address);

			// if write access to time control keep a copy of NTP mode bit
			if (address == TIMECTRL) {
				clock_m_dev->ntpModeEnabled = (data & TIMECTL_TIMEFORMAT) != 0;
			}
		} else {
			// write invalid virtual register address
			pr_err(
				 "unexpected write access to virtual register: 0x%lX\n",
				 address);
		}
		break;
	}

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_setreg);

bool clock_m_setreg_verify(struct clock_m_device *clock_m, unsigned long address, u32 data) {
	u32 tmp;
	clock_m_setreg(clock_m, address, data);
	clock_m_getreg(clock_m, address, &tmp);

	return data == tmp;

}
EXPORT_SYMBOL_GPL(clock_m_setreg_verify);


/**
 * <!--********************************************************************************-->
 * \defgroup clock_m_functions functions to control the clock synchronization cell
 * <!--********************************************************************************-->
 */

/**
 * Low level helper function for Lsync Utility.
 * Get System Time and 1588 Time with as few delay as possible
 * \param clock_m_dev pointer to clock_m_device structure
 * \param sys_ts system timestamp
 * \param hardware_ts 1588 hardware timestamp
 * \param delta_sys duration of getnstimeofday() in clock cycles
 * \param delta_hw duration of getting hardware time in clock cycles
 */
int clock_m_getTimestamps(struct clock_m_device *clock_m_dev, syn1588_timespec *sys_ts,
			  syn1588_timespec *hw_ts, unsigned long long *delta_sys,
			  unsigned long long *delta_hw)
{
	syn1588_timespec old_ts, ts;
	unsigned long long cycles1, cycles2, cycles3;
	u32 timev_m, timev_h;
	unsigned long flags;

	// we also disable all IRQs on the local cpu,
	// this also should disable preemption
	local_irq_save(flags);
	local_irq_disable();

	syn1588_ktime_get_real_ts(&old_ts);
	do {
		cycles1 = get_cycles();
		syn1588_ktime_get_real_ts(&ts);
		cycles2 = get_cycles();
	}
	// wait for nsec wrap, because even though it shows time in ns it only
	// has a resolution of 1 us.
	// we wait for the wrap and assume that the resulting nsec value is
	// one us with as few ns as possible (we cannot see those ns )
	while (ts.tv_nsec == old_ts.tv_nsec);

	// read hw time
	// this has to be done with as few instructions in between both calls
	// to get the smallest delta possible.

	timev_m = clock_m_readl(clock_m_dev, TIME2_M);
	cycles3 = get_cycles();
	local_irq_restore(flags);
	timev_h = clock_m_readl(clock_m_dev, TIME2_H);

	// copy everything over
	sys_ts->tv_nsec = ts.tv_nsec;
	sys_ts->tv_sec = ts.tv_sec;
	hw_ts->tv_nsec = timev_m;
	hw_ts->tv_sec = timev_h;

	// how many cycles for getnstimeofday?
	*delta_sys = cycles2 - cycles1;
	// how many cycles for hardware access?
	*delta_hw = cycles3 - cycles2;

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_getTimestamps);


/**
 * disable all clock synchronization cell interrupts of a specific adapter if
 * unused or forced
 * \param adapter pointer to syn1588(R) NIC device structure of the CLOCK_M the IRQs
 * should be disabled for
 * \param force if not set zero, disable all CLOCK_M IRQs, even if adapter still
 * seams to be used
 * \ingroup clock_m_functions
 */
void clock_m_dev_freeirq(struct clock_m_device *clock_m_dev)
{
	pr_info("disabling IRQ processing.\n");
	if(clock_m_dev->iobase != NULL) {
		clock_m_writel(clock_m_dev, 0x0, IREN); // disable all CLOCK_M IRQs
	}
}


/**
 * Write PHY delay values to the PHY delay registers
 *
 */

void set_mcore_phy_delay(struct clock_m_device *clock_m_dev, u32 tx_delay,
			 u32 rx_delay)
{
	int ret = 0;
	u32 reg = 0;
	// Set PHY delay registers
	reg = tx_delay;
	ret = clock_m_setreg(clock_m_dev, PHY_DELAY_TX, reg);

	reg = rx_delay;
	ret |= clock_m_setreg(clock_m_dev, PHY_DELAY_RX, reg);

	if (ret)
		pr_warn("PHY delay setting failed!\n");
}
EXPORT_SYMBOL_GPL(set_mcore_phy_delay);


static void clock_m_read_capabilities(struct clock_m_device *clock_m)
{
	u32 caps1 = clock_m_readl(clock_m, CAPABILITIES1);
	u32 caps2 = clock_m_readl(clock_m, CAPABILITIES2);
	//u32 caps2 = clock_m_readl(clock_m, CAPABILITIES3); // not used yet
	clock_m->clock_m_capabilities.miits0 = (caps1 &  BIT(0)) > 0;
	clock_m->clock_m_capabilities.miits1 = (caps1 &  BIT(1)) > 0;
	clock_m->clock_m_capabilities.lsync_sync_status = (caps1 &  BIT(2)) > 0;
	clock_m->clock_m_capabilities.measurement = (caps1 &  BIT(3)) > 0;
	clock_m->clock_m_capabilities.event0 = (caps1 &  BIT(4)) > 0;
	clock_m->clock_m_capabilities.event1 = (caps1 &  BIT(5)) > 0;
	clock_m->clock_m_capabilities.trigger0 = (caps1 &  BIT(6)) > 0;
	clock_m->clock_m_capabilities.trigger1 = (caps1 &  BIT(7)) > 0;
	clock_m->clock_m_capabilities.period0 = (caps1 &  BIT(8)) > 0;
	clock_m->clock_m_capabilities.period1 = (caps1 &  BIT(9)) > 0;
	clock_m->clock_m_capabilities.time2 = (caps1 &  BIT(10)) > 0;
	clock_m->clock_m_capabilities.irigb_output = (caps1 &  BIT(11)) > 0;
	clock_m->clock_m_capabilities.onestep = (caps1 &  BIT(12)) > 0;
	clock_m->clock_m_capabilities.period0cfg = (caps1 &  BIT(13)) > 0;
	clock_m->clock_m_capabilities.ntp = (caps1 &  BIT(14)) > 0;
	clock_m->clock_m_capabilities.leapset = (caps1 &  BIT(15)) > 0;
	clock_m->clock_m_capabilities.dutyperiod = (caps1 &  BIT(16)) > 0;
	clock_m->clock_m_capabilities.packetgenerator = (caps1 &  BIT(17)) > 0;
	clock_m->clock_m_capabilities.packetslaveformat = (caps1 &  BIT(18)) > 0;
	clock_m->clock_m_capabilities.precisionperiod = (caps1 &  BIT(19)) > 0;
	clock_m->clock_m_capabilities.irigb_input = (caps1 &  BIT(20)) > 0;
	clock_m->clock_m_capabilities.ext_vlan_pgkgen = (caps1 &  BIT(21)) > 0;
	clock_m->clock_m_capabilities.stop_clock = (caps1 &  BIT(22)) > 0;
	clock_m->clock_m_capabilities.hsnetworkindirect = (caps1 &  BIT(23)) > 0;
	clock_m->clock_m_capabilities.period2 = (caps1 &  BIT(24)) > 0;
	clock_m->clock_m_capabilities.period3 = (caps1 &  BIT(25)) > 0;
	clock_m->clock_m_capabilities.extime = (caps1 &  BIT(26)) > 0;
	clock_m->clock_m_capabilities.ptp_sync_status = (caps1 &  BIT(27)) > 0;
	clock_m->clock_m_capabilities.sdi_extended_support = (caps1 &  BIT(28)) > 0;
	clock_m->clock_m_capabilities.audio_counter = (caps1 &  BIT(29)) > 0;
	clock_m->clock_m_capabilities.sdi_ltc_timecode = (caps1 &  BIT(30)) > 0;

	clock_m->clock_m_capabilities.miits2_1g = (caps2 &  BIT(0)) > 0;
	clock_m->clock_m_capabilities.miits3_1g = (caps2 &  BIT(1)) > 0;
	clock_m->clock_m_capabilities.miits4_1g = (caps2 &  BIT(2)) > 0;
	clock_m->clock_m_capabilities.miits5_1g = (caps2 &  BIT(3)) > 0;
	clock_m->clock_m_capabilities.miits2_10g = (caps2 &  BIT(4)) > 0;
	clock_m->clock_m_capabilities.miits3_10g = (caps2 &  BIT(5)) > 0;
	clock_m->clock_m_capabilities.miits4_10g = (caps2 &  BIT(6)) > 0;
	clock_m->clock_m_capabilities.miits5_10g = (caps2 &  BIT(7)) > 0;
	clock_m->clock_m_capabilities.miits2_25g = (caps2 &  BIT(8)) > 0;
	clock_m->clock_m_capabilities.miits3_25g = (caps2 &  BIT(9)) > 0;
	clock_m->clock_m_capabilities.miits4_25g = (caps2 &  BIT(10)) > 0;
	clock_m->clock_m_capabilities.miits5_25g = (caps2 &  BIT(11)) > 0;
	clock_m->clock_m_capabilities.miits2_40g = (caps2 &  BIT(12)) > 0;
	clock_m->clock_m_capabilities.miits3_40g = (caps2 &  BIT(13)) > 0;
	clock_m->clock_m_capabilities.miits4_40g = (caps2 &  BIT(14)) > 0;
	clock_m->clock_m_capabilities.miits5_40g = (caps2 &  BIT(15)) > 0;
	clock_m->clock_m_capabilities.miits2_100g = (caps2 &  BIT(16)) > 0;
	clock_m->clock_m_capabilities.miits3_100g = (caps2 &  BIT(17)) > 0;
	clock_m->clock_m_capabilities.miits4_100g = (caps2 &  BIT(18)) > 0;
	clock_m->clock_m_capabilities.miits5_100g = (caps2 &  BIT(19)) > 0;
}

static void clock_m_print_capabilities(struct clock_m_device *clock_m)
{
	pr_info("clock_m miits0: %d\n", clock_m->clock_m_capabilities.miits0);
	pr_info("clock_m miits1: %d\n", clock_m->clock_m_capabilities.miits1);
	pr_info("clock_m lsync_sync_status: %d\n", clock_m->clock_m_capabilities.lsync_sync_status);
	pr_info("clock_m measurement: %d\n", clock_m->clock_m_capabilities.measurement);
	pr_info("clock_m event0: %d\n", clock_m->clock_m_capabilities.event0);
	pr_info("clock_m event1: %d\n", clock_m->clock_m_capabilities.event1);
	pr_info("clock_m trigger0: %d\n", clock_m->clock_m_capabilities.trigger0);
	pr_info("clock_m trigger1: %d\n", clock_m->clock_m_capabilities.trigger1);
	pr_info("clock_m period0: %d\n", clock_m->clock_m_capabilities.period0);
	pr_info("clock_m period1: %d\n", clock_m->clock_m_capabilities.period1);
	pr_info("clock_m time2: %d\n", clock_m->clock_m_capabilities.time2);
	pr_info("clock_m irogb_output: %d\n", clock_m->clock_m_capabilities.irigb_output);
	pr_info("clock_m onestep: %d\n", clock_m->clock_m_capabilities.onestep);
	pr_info("clock_m period0cfg: %d\n", clock_m->clock_m_capabilities.period0cfg);
	pr_info("clock_m ntp: %d\n", clock_m->clock_m_capabilities.ntp);
	pr_info("clock_m leapset: %d\n", clock_m->clock_m_capabilities.leapset);
	pr_info("clock_m dutyperiod: %d\n", clock_m->clock_m_capabilities.dutyperiod);
	pr_info("clock_m packetgenerator: %d\n", clock_m->clock_m_capabilities.packetgenerator);
	pr_info("clock_m packetslaveformat: %d\n", clock_m->clock_m_capabilities.packetslaveformat);
	pr_info("clock_m precisionperiod: %d\n", clock_m->clock_m_capabilities.precisionperiod);
	pr_info("clock_m irigb_input: %d\n", clock_m->clock_m_capabilities.irigb_input);
	pr_info("clock_m ext_vlan_pgkgen: %d\n", clock_m->clock_m_capabilities.ext_vlan_pgkgen);
	pr_info("clock_m stop_clock: %d\n", clock_m->clock_m_capabilities.stop_clock);
	pr_info("clock_m hsnetworkindirect: %d\n", clock_m->clock_m_capabilities.hsnetworkindirect);
	pr_info("clock_m period2: %d\n", clock_m->clock_m_capabilities.period2);
	pr_info("clock_m period3: %d\n", clock_m->clock_m_capabilities.period3);
	pr_info("clock_m extime: %d\n", clock_m->clock_m_capabilities.extime);
	pr_info("clock_m ptp_sync_status: %d\n", clock_m->clock_m_capabilities.ptp_sync_status);
	pr_info("clock_m sdi_extended_support: %d\n", clock_m->clock_m_capabilities.sdi_extended_support);
	pr_info("clock_m audio_counter: %d\n", clock_m->clock_m_capabilities.audio_counter);
	pr_info("clock_m sdi_ltc_timecode: %d\n", clock_m->clock_m_capabilities.sdi_ltc_timecode);

	pr_info("clock_m miits2_1g: %d\n", clock_m->clock_m_capabilities.miits2_1g);
	pr_info("clock_m miits3_1g: %d\n", clock_m->clock_m_capabilities.miits3_1g);
	pr_info("clock_m miits4_1g: %d\n", clock_m->clock_m_capabilities.miits4_1g);
	pr_info("clock_m miits5_1g: %d\n", clock_m->clock_m_capabilities.miits5_1g);
	pr_info("clock_m miits2_10g: %d\n", clock_m->clock_m_capabilities.miits2_10g);
	pr_info("clock_m miits3_10g: %d\n", clock_m->clock_m_capabilities.miits3_10g);
	pr_info("clock_m miits4_10g: %d\n", clock_m->clock_m_capabilities.miits4_10g);
	pr_info("clock_m miits5_10g: %d\n", clock_m->clock_m_capabilities.miits5_10g);
	pr_info("clock_m miits2_25g: %d\n", clock_m->clock_m_capabilities.miits2_25g);
	pr_info("clock_m miits3_25g: %d\n", clock_m->clock_m_capabilities.miits3_25g);
	pr_info("clock_m miits4_25g: %d\n", clock_m->clock_m_capabilities.miits4_25g);
	pr_info("clock_m miits5_25g: %d\n", clock_m->clock_m_capabilities.miits5_25g);
	pr_info("clock_m miits2_40g: %d\n", clock_m->clock_m_capabilities.miits2_40g);
	pr_info("clock_m miits3_40g: %d\n", clock_m->clock_m_capabilities.miits3_40g);
	pr_info("clock_m miits4_40g: %d\n", clock_m->clock_m_capabilities.miits4_40g);
	pr_info("clock_m miits5_40g: %d\n", clock_m->clock_m_capabilities.miits5_40g);
	pr_info("clock_m miits2_100g: %d\n", clock_m->clock_m_capabilities.miits2_100g);
	pr_info("clock_m miits3_100g: %d\n", clock_m->clock_m_capabilities.miits3_100g);
	pr_info("clock_m miits4_100g: %d\n", clock_m->clock_m_capabilities.miits4_100g);
	pr_info("clock_m miits5_100g: %d\n", clock_m->clock_m_capabilities.miits5_100g);
}

static u64 div_rem(u64 n, u32 base, u32 *remainder){
	*remainder = do_div(n, base);
	return n;
}

/**
 * Hardware wants a 40bit fractal of ns
 */
static u64 calculate_step_register(u32 frequency) {
	u32 remainder;
	u64 step_ns = div_rem(NSEC_PER_SEC, frequency,  &remainder);

	u64 tmp = div_rem(((u64)remainder) << 32, frequency,  &remainder);
	u64 tmp2 = div_rem(((u64)remainder) << 8, frequency,  &remainder);
	u64 step_fractal = (step_ns << 40) + (tmp << 8) + tmp2;

	return step_fractal;
}


irqreturn_t clock_m_interrupt(int irq, void *data);

/**
 * Function, that initializes the clock synchronization core
 * steps performed: allocate data structure, measure clock frequency, set
 * initial step size
 * \param adapter pointer to a syn1588(R) NIC card data structure
 * \ingroup clock_m_functions
 */


static struct clock_m_device *clock_m_dev_init(struct clock_m_device *clock_m, int irq)
{
	u32 version_reg;
	int ret;
	u64 step_reg;
	struct device *dev = clock_m->dev;

	pr_debug("*** clock_m_dev_init ***\n");
	// initialize CLOCK_M device
	//clock_m->adapter = adapter; // store pointer to card
	clock_m->lost_irqs = 0; // init lost interrupt mask

	// initialize spin locks
	spin_lock_init(&clock_m->reg_lock);
	spin_lock_init(&clock_m->reg_lock_eventctrl);
	spin_lock_init(&clock_m->reg_lock_iomatrix);
	spin_lock_init(&clock_m->time1_lock);
	spin_lock_init(&clock_m->time2_lock);

	version_reg = ioread32(clock_m->iobase + SYN1588_VER);
	clock_m->version[0] = version_reg >> 24 & 0xff;
	clock_m->version[1] = version_reg >> 16 & 0xff;
	clock_m->version[2] = version_reg >> 8  & 0xff;
	clock_m->version[3] = version_reg >> 0  & 0xff;
	clock_m->version[4] = 0;

	dev_info(dev, "detected syn1588(R) Clock version %s.\n",
		 clock_m->version);

	if (clock_m->version[0] != 'L' && clock_m->version[0] != 'M') // check on leading L or M
	{
		dev_err(dev, "unable to retrieve syn1588(R) CLOCK_M version info!\n");
		return ERR_PTR(-ENODEV);
	}

	// check capability register
	clock_m_read_capabilities(clock_m);
	clock_m_print_capabilities(clock_m);

	clock_m->clk_freq = clock_m_readl(clock_m, CLK_FREQ);


	dev_info(dev, "preparing device file.\n");

	// calculate initial step size
	if ((clock_m->clk_freq < 1000000L) ||
		(clock_m->clk_freq > 1000000000L)) {
		pr_err(
			 "configured clock frequency (%llu kHz) out of range [1-1000 MHz]\n",
			 div64_u64(clock_m->clk_freq, 1000L));
		return ERR_PTR(-EOPNOTSUPP);
	}

	dev_info(dev, "configured clock frequency: %llu kHz.\n",
		 div64_u64(clock_m->clk_freq, 1000L));


	if(syn1588_disabled_output) {
		u32 tmp;
		spin_lock(&clock_m->reg_lock_iomatrix);
		tmp = clock_m_readl(clock_m, IOMATRIX);
		pr_debug("disable io matrix outputs by default\n");
		clock_m_writel(clock_m, tmp & 0xffff0000, IOMATRIX);
		spin_unlock(&clock_m->reg_lock_iomatrix);
	}

	step_reg = calculate_step_register(clock_m->clk_freq);
	dev_info(dev, "setting initial clock step size to %llu.%03u ns.\n",
		 step_reg >> 40, ((u32)((step_reg*1000) >> 40 ) ) % 1000  );
	// apply changes to CLOCK_M
	// set frac. step size
	clock_m_writel(clock_m, (u32)step_reg, SHDWSTEPPU_L);
	// set step size
	clock_m_writel(clock_m, (u32)(step_reg >> 32), SHDWSTEPPU_H);
	// write step size
	clock_m_writel(clock_m, TIMECTL_LOADSTEP, TIMECTRL);
	// reset time control register
	clock_m_writel(clock_m, 0x0, TIMECTRL);


	ret = syn1588nic_phc_init(clock_m);
	if(ret != 0) {
		dev_err(dev, "could not init phc\n");
		return ERR_PTR(ret);
	}

	ret = clock_m_io_probe(clock_m);
	if(ret != 0) {
		dev_err(dev, "could not init io part\n");
		return ERR_PTR(ret);
	}

	tasklet_init(&clock_m->tl_clock_m, syn1588_BHH_CLOCK_M, (unsigned long)clock_m);
	if(irq > 0) {
		dev_dbg(dev, "request irq #%d for %p\n", irq, clock_m);
		ret = request_irq(irq, &clock_m_interrupt,
					IRQF_DISABLED | IRQF_SHARED, DRV_NAME,
					clock_m);
		if (ret > 0) {
			dev_err(dev, "Could not obtain IRQ #%u\n", irq);
			return ERR_PTR(-EIO);
		}
		clock_m->irq = irq;
	} else {
		dev_warn(dev, "no interrupt support");
	}
	return clock_m;
}

int clock_m_register_callback(struct clock_m_device *clock_m, struct oregano_clock_callback *callback) {
	if(clock_m->callback != NULL) {
		return -ENOMEM;
	}
	clock_m->callback = callback;
	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_register_callback);

int clock_m_unregister_callback(struct clock_m_device *clock_m, struct oregano_clock_callback *callback){
	if(clock_m->callback == callback) {
		clock_m->callback = NULL;
		return 0;
	}

	return -ENODEV;
}
EXPORT_SYMBOL_GPL(clock_m_unregister_callback);

irqreturn_t clock_m_interrupt(int irq, void *data)
{
	struct clock_m_device *clock_m_device = data;
	u32 temp_clock_m_irsrc;
	int count;

	temp_clock_m_irsrc =
		clock_m_readl(clock_m_device, IRSRC); // read CLOCK_M interrupt source register

	if (!temp_clock_m_irsrc)
		return IRQ_NONE; // this call was due to a shared interrupt line

	if (temp_clock_m_irsrc != 0) {
#ifndef SYN1588Clock_L
		// clear interrupt source register (acknowledge read irq's)
		clock_m_writel(clock_m_device, ~temp_clock_m_irsrc, IRSRC);
#endif

		// copy interrupts that were read out
		for (count = 0; count < sizeof(u32) * BITS_PER_BYTE; count++)
			if (unlikely((0 != (temp_clock_m_irsrc & 0x1 << count)) &&
					 (0 !=
					  test_and_set_bit(count,
							   &clock_m_device->irh_irsrc))))
			{
				set_bit(count, &clock_m_device->lost_irqs);
			}

		tasklet_schedule(&clock_m_device->tl_clock_m);
	}

	return IRQ_HANDLED;
}



/**
 * Bottom half IRQ handler
 */
static void syn1588_BHH_CLOCK_M(unsigned long pt_adapter)
{
	struct clock_m_device *clock_m = (struct clock_m_device *)pt_adapter;
	struct device *dev = clock_m->dev;
	unsigned char failure;
	int count;

	if(test_and_clear_bit(CLOCK_M_IRSC_EVENT0OCCURED, &clock_m->irh_irsrc))
	{
		if(clock_m_event_read_ts(clock_m, SYN1588_IOCTL_EVENT0) == -EINVAL) {
			if(test_and_set_bit(CLOCK_M_IRSC_EVENT0OCCURED, &clock_m->bhh_irsrc)) {
				dev_crit(dev, "BH handler syn1588(R) Clock IRQ EVENT0 handling failure!\n");
			}
		}
	}

	if(test_and_clear_bit(CLOCK_M_IRSC_EVENT1OCCURED, &clock_m->irh_irsrc))
	{
		if(clock_m_event_read_ts(clock_m, SYN1588_IOCTL_EVENT1) == -EINVAL) {
			if(test_and_set_bit(CLOCK_M_IRSC_EVENT1OCCURED, &clock_m->bhh_irsrc)) {
				dev_crit(dev, "BH handler syn1588(R) Clock IRQ EVENT1 handling failure!\n");
			}
		}
	}

	if(test_and_clear_bit(CLOCK_M_IRSC_TRIGGER0MATCH, &clock_m->irh_irsrc))
	{
		if(clock_m_trigger_write_ts(clock_m, SYN1588_IOCTL_TRIGGER0) == -EINVAL) {
			if(test_and_set_bit(CLOCK_M_IRSC_TRIGGER0MATCH, &clock_m->bhh_irsrc)) {
				dev_crit(dev, "BH handler syn1588(R) Clock IRQ TRIGGER0 handling failure!\n");
			}
		}
	}

	if(test_and_clear_bit(CLOCK_M_IRSC_TRIGGER1MATCH, &clock_m->irh_irsrc))
	{
		if(clock_m_trigger_write_ts(clock_m, SYN1588_IOCTL_TRIGGER1) == -EINVAL) {
			if(test_and_set_bit(CLOCK_M_IRSC_TRIGGER1MATCH, &clock_m->bhh_irsrc)) {
				dev_crit(dev, "BH handler syn1588(R) Clock IRQ TRIGGER1 handling failure!\n");
			}
		}
	}
	if(!clock_m->irh_irsrc && !clock_m->bhh_irsrc) {
		// no depricated irqs to handel
		return ;
	}

	// readout and acknowledge CLOCK_M interrupts
	failure = 0;
	for (count = 0; count < sizeof(u32) * BITS_PER_BYTE; count++)
	{
		if (unlikely(0 != (test_and_clear_bit(count,
							  &clock_m->irh_irsrc)) &&
				 (0 !=
				  test_and_set_bit(count, &clock_m->bhh_irsrc))))
		{
			failure = 1;
		}
	}
	if (unlikely(failure))
		pr_crit("BH handler syn1588(R) Clock IRQ handling failure!\n");

	pr_debug("*** BH handler (mask Clock 0x%lX) ***\n",
		 clock_m->bhh_irsrc);
	if (unlikely(clock_m->lost_irqs != 0)) {
		pr_warn("bottom-half handler lost interrupt (mask Clock 0x%lX).\n",
			 clock_m->lost_irqs);
		for (count = 0; count < sizeof(u32) * BITS_PER_BYTE; count++) {
			clear_bit(count, &clock_m->lost_irqs);
		}
	}

	if ((!clock_m->bhh_irsrc)) {
		pr_info("IRQ BHH called with empty IRSRC!\n");
		return;
	}

	if (clock_m->bhh_irsrc) {
		struct oregano_clock_event_data event  = {};
		event.irs = clock_m->bhh_irsrc;
		if(clock_m->callback != NULL) {
			clock_m->callback->fn(clock_m, &event, clock_m->callback->data );
			clock_m->bhh_irsrc = 0;
		} else {
			dev_err(dev, "internal error - not clock_m callback set! \n");
		}
	}
}




/// @{ paramter parsing
static int oregano_clock_probe_pdata(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct clock_m_device *clock_m = platform_get_drvdata(pdev);
	struct oregano_clock_platform_data *pdata = dev_get_platdata(dev);

	clock_m->si_time_pll = pdata->si_time_pll;

	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id oregano_clock_dt_ids[] = {
	{ .compatible = DRV_NAME },
	{ .compatible = "oregano,clock_m" },
	{}
};
MODULE_DEVICE_TABLE(of, oregano_clock_dt_ids);

static int oregano_clock_probe_dt(struct platform_device *pdev)
{

	struct device *dev = &pdev->dev;
	struct clock_m_device *clock_m = platform_get_drvdata(pdev);

	struct device_node *pll_node;


	pll_node = of_parse_phandle(pdev->dev.of_node, "pll", 0);
	if(!IS_ERR_OR_NULL(pll_node)) {
		struct device *pll_dev = bus_find_device_by_of_node(&i2c_bus_type, pll_node);
		if (IS_ERR_OR_NULL(pll_dev)) {
			dev_err(dev, "of_find_device_by_node(%pOF) failed, %p\n", pll_node, pll_dev);
			return -ENODEV;
		}
		clock_m->si_time_pll = sit_95147_get_device(pll_dev);
	}

	return 0;
}
#else
static inline int oregano_clock_probe_dt(struct platform_device *pdev)
{
	return 0;
}
#endif

/// @} paramter parsing


static int oregano_clock_remove(struct platform_device *pdev)
{
	struct clock_m_device *clock_m = platform_get_drvdata(pdev);
	struct device *dev = &pdev->dev;

	dev_info(dev, "exiting driver");
	if(IS_ERR_OR_NULL(clock_m)) {
		return 0;
	}
	dev_dbg(dev, "unregistering syn1588 device.\n");
	clock_m_dev_freeirq(clock_m); // turn off CLOCK_M IRQ's
	if(clock_m->irq > 0) {
		pr_debug("free irq #%d for %p\n", clock_m->irq, clock_m);
		free_irq(clock_m->irq, clock_m);
		clock_m->irq = 0;
	}
	tasklet_kill(&clock_m->tl_clock_m); // kill bottom half CLOCK_M handler
	syn1588nic_phc_exit(clock_m);

	kfree(clock_m);
	platform_set_drvdata(pdev, NULL);

	return 0;
}

static int oregano_clock_probe(struct platform_device *pdev)
{
	int				status;
	struct device			*dev = &pdev->dev;
	struct resource *res;
	struct clock_m_device *clock_m = kzalloc(sizeof(struct clock_m_device), GFP_KERNEL);
	int irq = platform_get_irq_byname(pdev, "main");
	if(irq < 0) {
		irq = platform_get_irq(pdev, 0);
	}
	platform_set_drvdata(pdev, clock_m);
	clock_m->dev = &pdev->dev;

	dev_info(dev, "probe oregano clock driver\n");

	if (dev_of_node(&pdev->dev))
		status = oregano_clock_probe_dt(pdev);
	else
		status = oregano_clock_probe_pdata(pdev);

	if (status)
		return status;

	if(irq <=0) {
		dev_err(dev, "no valid interrupt provided\n");
		return -ENODEV;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if(res == NULL) {
		dev_err(dev, "could not request resources\n");
		return -ENODEV;
	}

	clock_m->iobase = devm_ioremap_resource(dev, res);
	if(IS_ERR_OR_NULL(clock_m->iobase)) {
		dev_err(dev, "could not ioremap resource\n");
		return PTR_ERR(clock_m->iobase);
	}
	clock_m->resource = res;


	clock_m_dev_init(clock_m, irq);
	return 0;
}








MODULE_ALIAS("platform:" DRV_NAME);

static struct platform_driver oregano_clock_driver = {
	.driver = {
		.name	= DRV_NAME,
#ifdef CONFIG_OF
		.of_match_table = of_match_ptr(oregano_clock_dt_ids),
#endif // CONFIG_OF
	},
	.probe		= oregano_clock_probe,
	.remove		 = oregano_clock_remove,
};
module_platform_driver(oregano_clock_driver);

MODULE_DESCRIPTION("clock driver for oregano");
MODULE_AUTHOR("Norbert Summer");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);
