// SPDX-License-Identifier: GPL-2.0
//
// Driver for SI-Time 95147
//
// Copyright (C) 2023 Oregano Systems
//
// Author: Norbert Summer <summer@oregano.at>

#include "si_time_pll.h"
#include "compat/compat.h"
#include "version.h"

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>
#include <linux/mutex.h>


struct si_time_95147_spi {
	struct spi_device *spi;
};
struct si_time_95147_i2c {
	struct i2c_client *i2c;
};

enum pll_bus_type {
	PLL_BUS_TYPE_SPI,
	PLL_BUS_TYPE_I2C,
};

struct si_time_95147 {
	struct device *dev;
	enum pll_bus_type bus_type;
	union {
		struct si_time_95147_spi spi;
		struct si_time_95147_i2c i2c;
	} bus;

	struct mutex lock;
	struct work_struct	adjfreq_work;

	s64 current_adjfreq; //scaled_ppm
	s64 pending_adjfreq; //scaled_ppm


	int clock_ajust_page; // 0xa-0xd
	int clock_input_pll_page; // 0xa-0xd
	int clock_input; // 0-3
	int external_clock_input; // 1-3
	int sync_e_clock_input; // 1-3
	int sync_e_pll_page; // 0xa-0xd or -1

	struct si_time_pll_status status;

	// paramter for clock adjust, read from the pll
	u32 divn;
	u32 divn2_int;
	s32 divn2_fracnum;
	u32 divn2_fracden;

	struct timer_list watchdog; // timeout to poll pll status
	struct work_struct watchdog_work;
};

struct si_time_pll_internal_status {
	int id;
	bool loss_of_lock;
	bool loss_of_lock_sticky;
	bool fast_lock_mode;
	bool holdover_mode; // equals to free run mode
	bool holdover_mode_sticky;
	bool holdover_valid;
};

struct si_time_pll_internal_input_status {
	bool loss_of_lock;
	bool loss_of_lock_sticky;
};

enum sit_95147_reg_page0 {
	SIT_95147_PAGE0_VERSION_ID = 0x01,
	SIT_95147_PAGE0_PAGE_NUMER = 0xff,
};

enum sit_95147_reg_page1 {
	SIT_95147_PAGE1_STATUS_CLOCKMON = 0x02,
	SIT_95147_PAGE1_NOTIFY_CLOCKMON = 0x03,
};

enum sit_95147_reg_page_pll {
	SIT_95147_PAGE_PLL_STATUS_PLL = 0x02,
	SIT_95147_PAGE_PLL_NOTIFY_PLL = 0x03,
  	SIT_95147_PAGE_PLL_STATUS1_PLL = 0x06,
	SIT_95147_PAGE_PLL_NOTIFY1_PLL = 0x07,
	SIT_95147_PAGE_PLL_DIVNINT = 0x17,
	SIT_95147_PAGE_PLL_MISCXO = 0x18,
	SIT_95147_PAGE_PLL_DIVNFRAC2 = 0x1a,
	SIT_95147_PAGE_PLL_DIVNFRAC3 = 0x1b,
	SIT_95147_PAGE_PLL_DIVNFRAC4 = 0x1c,
	SIT_95147_PAGE_PLL_DIVN2_INT1 = 0x1d,
	SIT_95147_PAGE_PLL_DIVN2_INT2 = 0x1e,
	SIT_95147_PAGE_PLL_DIVN2_INT3 = 0x1f,
	SIT_95147_PAGE_PLL_DIVN2_FRN1 = 0x20,
	SIT_95147_PAGE_PLL_DIVN2_FRN2 = 0x21,
	SIT_95147_PAGE_PLL_DIVN2_FRN3 = 0x22,
	SIT_95147_PAGE_PLL_DIVN2_FRN4 = 0x23,
	SIT_95147_PAGE_PLL_DIVN2_FRD1 = 0x24,
	SIT_95147_PAGE_PLL_DIVN2_FRD2 = 0x25,
	SIT_95147_PAGE_PLL_DIVN2_FRD3 = 0x26,
	SIT_95147_PAGE_PLL_DIVN2_FRD4 = 0x27,
	SIT_95147_PAGE_PLL_DCO_FRAC1 = 0x31,
	SIT_95147_PAGE_PLL_DCO_FRAC2 = 0x32,
	SIT_95147_PAGE_PLL_DCO_FRAC3 = 0x33,
	SIT_95147_PAGE_PLL_DCO_FRAC4 = 0x34,
	SIT_95147_PAGE_PLL_DCO_FUNCTION = 0x35,
	SIT_95147_PAGE_PLL_DCO_BUMP2 = 0x36,
	SIT_95147_PAGE_PLL_DCO_BUMP3 = 0x37,
};

enum sit_95147_spi_mode {
	SIT_95147_SPI_MODE_READ = 0x00,
	SIT_95147_SPI_MODE_WRITE = 0x01,
};

static int sit_95147_write_reg(struct si_time_95147 *pll, u8 reg, u8 value) {
	if (pll->bus_type == PLL_BUS_TYPE_SPI) {

		struct spi_message m;
		u8 buf[] = {SIT_95147_SPI_MODE_WRITE, reg, value};
		struct spi_transfer t = {
			.tx_buf = &buf,
			.len = ARRAY_SIZE(buf),
		};

		spi_message_init(&m);
		spi_message_add_tail(&t, &m);
		return spi_sync(pll->bus.spi.spi, &m);
	} else if (pll->bus_type == PLL_BUS_TYPE_I2C) {
		u8 buf[2] = {reg, value};
		return i2c_master_send(pll->bus.i2c.i2c, (char *)&buf, 2);
	} else {
		return -ENOTSUPP;
	}
}

static u8 sit_95147_read_reg(struct si_time_95147 *pll, u8 reg) {
	if (pll->bus_type == PLL_BUS_TYPE_SPI) {

		struct spi_message m;
		u8 buf_tx[3] = {SIT_95147_SPI_MODE_READ, reg, 0x00};
		u8 buf_rx[3] = {};
		struct spi_transfer t = {
			.tx_buf = &buf_tx,
			.len = ARRAY_SIZE(buf_tx),
			.rx_buf = &buf_rx,
		};
		int ret;

		spi_message_init(&m);
		spi_message_add_tail(&t, &m);
		ret = spi_sync(pll->bus.spi.spi, &m);
		if(!ret) {
			return buf_rx[2];
		}
		pr_err("error reading spi regiser");
		return 0;
	} else if (pll->bus_type == PLL_BUS_TYPE_I2C) {
		u8 value;
		int result = i2c_master_send(pll->bus.i2c.i2c, (char *)&reg, 1);
		if(!result) {
			pr_err("error reading i2c regiser");
			return 0;
		}
		result = i2c_master_recv(pll->bus.i2c.i2c, &value, 1);
		if(!result) {
			pr_err("error reading i2c regiser");
			return 0;
		}
		return value;
	} else {
		pr_err("internal error, unkown bus type");
		return 0;
		//return -ENOTSUPP;
	}
}

static int sit_95147_set_page(struct si_time_95147 *pll, int page) {
	return sit_95147_write_reg(pll, SIT_95147_PAGE0_PAGE_NUMER, page);
}

static void si_time_pll_clock_adjfine_work(struct work_struct *work);

static int sit_95147_probe_pdata(struct si_time_95147 *pll, const struct si_time_pll_platform_data *platform_data) {
	struct device *dev = pll->dev;

	if (platform_data->clock_input_pll < 'A' || platform_data->clock_input_pll > 'D') {
		dev_err(dev, "invalid clock_input_pll definded '%c' is not A..D\n", platform_data->clock_input_pll);
		return -EINVAL;
	}
	pll->clock_input_pll_page = platform_data->clock_input_pll - 'A' + 0xa;

	if (platform_data->clock_adjust_pll < 'A' || platform_data->clock_adjust_pll > 'D') {
		dev_err(dev, "invalid clock_adjust_pll definded '%c' is not A..D\n", platform_data->clock_adjust_pll);
		return -EINVAL;
	}
	pll->clock_ajust_page = platform_data->clock_adjust_pll - 'A' + 0xa;

	if (platform_data->clock_input < 0 || platform_data->clock_input > 3) {
		dev_err(dev, "invalid clock_input definded '%d' is not 0-3\n", platform_data->clock_input);
		return -EINVAL;
	}
	pll->clock_input = platform_data->clock_input;


	if (platform_data->external_clock_input >= 0 || platform_data->external_clock_input <= 3) {
		pll->external_clock_input = platform_data->external_clock_input;
	} else {
		pll->external_clock_input = -1;
	}

	if (platform_data->sync_e_clock_input >= 0 && platform_data->sync_e_clock_input <= 3) {
		pll->sync_e_clock_input = platform_data->sync_e_clock_input;
	} else {
		pll->sync_e_clock_input = -1;
	}

	if (platform_data->sync_e_pll >= 'A' && platform_data->sync_e_pll <= 'D') {
		pll->sync_e_pll_page = platform_data->sync_e_pll - 'A' + 0xa;
	} else {
		pll->sync_e_pll_page = -1;
	}

	return 0;
}

#ifdef CONFIG_OF

static int sit_95147_probe_dt(struct si_time_95147 *pll) {
	struct device *dev = pll->dev;

	int ret;
	const char *clock_adjust_pll_str = NULL;
	const char *clock_input_pll_str = NULL;
	const char *sync_e_pll_str = NULL;
	struct si_time_pll_platform_data platform_data = {};
	u8 clock_input;
	u8 external_clock_input;
	u8 sync_e_clock_input;


	ret = of_property_read_string(dev->of_node, "clock_adjust_pll", &clock_adjust_pll_str);
	if (ret) {
		dev_err(dev, "of_property_read_string(clock_adjust_pll) failed with (%d)\n", ret);
		return -ENODEV;
	}
	if(clock_adjust_pll_str == NULL) {
		dev_err(dev, "of_property_read_string(clock_adjust_pll) string is NULL\n");
		return -ENODEV;
	}
	if(strlen(clock_adjust_pll_str) != 1) {
		dev_err(dev, "clock_adjust_pll_str is not length 1\n");
		return -ENODEV;
	}
	platform_data.clock_adjust_pll = clock_adjust_pll_str[0];

	ret = of_property_read_string(dev->of_node, "clock_input_pll", &clock_input_pll_str);
	if (ret) {
		dev_err(dev, "of_property_read_string(clock_input_pll) failed with (%d)\n", ret);
		return -ENODEV;
	}
	if(clock_input_pll_str == NULL) {
		dev_err(dev, "of_property_read_string(clock_input_pll) string is NULL\n");
		return -ENODEV;
	}
	if(strlen(clock_input_pll_str) != 1) {
		dev_err(dev, "clock_adjust_pll_str is not length 1\n");
		return -ENODEV;
	}
	platform_data.clock_input_pll = clock_input_pll_str[0];


	platform_data.clock_input = -1;
	ret = of_property_read_u8(dev->of_node, "clock_input", &clock_input);
	if (ret) {
		dev_err(dev, "of_property_read_u8(clock_input) failed with (%d)\n", ret);
		return -ENODEV;
	}
	platform_data.clock_input = clock_input;


	platform_data.external_clock_input = -1;
	ret = of_property_read_u8(dev->of_node, "external_clock_input", &external_clock_input);
	if (ret == 0) {
		platform_data.external_clock_input = external_clock_input;
	}

	platform_data.sync_e_clock_input = -1;
	ret = of_property_read_u8(dev->of_node, "sync_e_clock_input", &sync_e_clock_input);
	if (ret == 0) {
		platform_data.sync_e_clock_input = sync_e_clock_input;
	}


	ret = of_property_read_string(dev->of_node, "sync_e_pll", &sync_e_pll_str);
	if(ret == 0 && sync_e_pll_str != NULL) {
		if(strlen(sync_e_pll_str) != 1) {
			dev_err(dev, "sync_e_pll_str is not length 1\n");
			return -ENODEV;
		}

		platform_data.sync_e_pll = sync_e_pll_str[0];
	}

	return sit_95147_probe_pdata(pll, &platform_data) ;
}
#else
static int sit_95147_probe_dt(struct si_time_95147 *pll)
{
	return 0;
}
#endif


static void si_time_pll_get_status_internal(struct si_time_95147 *pll, int pll_page, struct si_time_pll_internal_status *status) {
	u32 tmp;
	sit_95147_set_page(pll, pll_page);
	status->id = pll_page;

	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_STATUS_PLL);
	status->loss_of_lock = (tmp & 1u) > 0;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_NOTIFY_PLL);
	status->loss_of_lock_sticky = (tmp & 1u) > 0;
	sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_NOTIFY_PLL, tmp);

	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_STATUS1_PLL);
	status->holdover_mode = (tmp & 1u) > 0;
	status->fast_lock_mode = (tmp & (1u<<1)) > 0;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_NOTIFY1_PLL);
	status->holdover_mode_sticky = (tmp & 1u) > 0;
	sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_NOTIFY1_PLL, tmp);

	//TODO
	status->holdover_valid = false;
}
static void si_time_pll_input_get_status_internal(struct si_time_95147 *pll, int input_nr, struct si_time_pll_internal_input_status *status) {
	u32 tmp;
	sit_95147_set_page(pll, 0x01);

	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE1_STATUS_CLOCKMON);
	status->loss_of_lock = (tmp & (1u << input_nr) ) != 0;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE1_NOTIFY_CLOCKMON);
	status->loss_of_lock_sticky = (tmp & (1u << input_nr) ) != 0;
	sit_95147_write_reg(pll, SIT_95147_PAGE1_NOTIFY_CLOCKMON, tmp);
}


// define poll time to 2s
#define POLL_TIME (2 * HZ)

static void sit_95147_poll_status(struct work_struct *work) {
	struct si_time_95147 *pll = container_of(work, struct si_time_95147, watchdog_work);
	struct device *dev = pll->dev;
	struct si_time_pll_status si_time_status;
	struct si_time_pll_internal_status status;
	struct si_time_pll_internal_input_status status_input;

	mutex_lock(&pll->lock);

	si_time_pll_get_status_internal(pll, pll->clock_input_pll_page, &status);
	si_time_status.input_pll_locked = true;
	if(status.loss_of_lock || status.loss_of_lock_sticky) {
		dev_err(dev, "!!!lost main clock input lock in pll!!!\n");
		dev_err(dev, "status: LoL: %d, LoL-sticky: %d\n", status.loss_of_lock, status.loss_of_lock_sticky);
		si_time_status.input_pll_locked = false;
	}

	si_time_pll_input_get_status_internal(pll, pll->clock_input, &status_input);
	si_time_status.input_locked = true;
	if(status_input.loss_of_lock || status_input.loss_of_lock_sticky) {
		dev_err(dev, "!!!lost main clock input lock!!!\n");
		dev_err(dev, "status: LoL: %d, LoL-sticky: %d\n", status_input.loss_of_lock, status_input.loss_of_lock_sticky);
		si_time_status.input_locked = false;
	}


	si_time_status.extclock_input_locked = false;
	if(pll->external_clock_input >= 0) {
		si_time_pll_input_get_status_internal(pll, pll->external_clock_input, &status_input);
		if(!status_input.loss_of_lock) {
			dev_err(dev, "We have external clock input\n");
		}
		si_time_status.extclock_input_locked = true;
	}

	si_time_status.synce_input_locked = false;
	if(pll->sync_e_clock_input >= 0) {
		si_time_pll_input_get_status_internal(pll, pll->sync_e_clock_input, &status_input);
		if(!status_input.loss_of_lock) {
			dev_dbg(dev, "We have sync-e clock input\n");
		}
		si_time_status.synce_input_locked = true;
	}
	if(pll->sync_e_pll_page >= 0) {
		si_time_pll_get_status_internal(pll, pll->sync_e_pll_page, &status);
		if(!status.loss_of_lock) {
			dev_err(dev, "We have sync-e pll lock\n");
		}
	}

	mutex_unlock(&pll->lock);

	pll->status = si_time_status;

}

static void sit_95147_poll_status_timer(struct timer_list *timer) {
	struct si_time_95147 *pll = from_timer(pll, timer, watchdog);

	pll->watchdog.expires = jiffies + POLL_TIME;
	add_timer(&pll->watchdog);

	//schedule async work without atomic context
	schedule_work(&pll->watchdog_work);
}

static ssize_t si_time_pll_show_clock_input(struct device *dev, struct device_attribute *attr,
                        char *buf)
{
	struct si_time_95147* pll = dev_get_drvdata(dev);

        return snprintf(buf, PAGE_SIZE, "%d\n", pll->status.input_locked);
}

static DEVICE_ATTR(clock_input, S_IRUGO, si_time_pll_show_clock_input, NULL);

void sit_95147_exit(struct si_time_95147 *pll);

static int sit_95147_probe(struct si_time_95147 *pll) {
	struct device *dev = pll->dev;
	u8 version;
	u8 tmp;
	u32 divn2_int;
	s32 divn2_fracnum;
	u32 divn2_fracden;
	u64 divn;
	int status;


	if (dev_of_node(dev))
		status = sit_95147_probe_dt(pll);
	else
		status = sit_95147_probe_pdata(pll, dev->platform_data);

	if (status)
		return status;

	mutex_init(&pll->lock);


	sit_95147_set_page(pll, 0x00);
	version = sit_95147_read_reg(pll, SIT_95147_PAGE0_VERSION_ID);

	if(version != 0xC1 && version != 0xC2) {
		dev_err(dev, "read invalid version number, got 0x%02x\n", version);
		return -ENODEV;
	}

	sit_95147_set_page(pll, pll->clock_ajust_page);

	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_INT1);
	divn2_int = tmp;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_INT2);
	divn2_int |= (u32)tmp << 8;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_INT3);
	divn2_int |= (u32)(tmp & 0x1f) << 16;


	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_FRN1);
	divn2_fracnum = tmp;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_FRN2);
	divn2_fracnum |= (u32)tmp << 8;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_FRN3);
	divn2_fracnum |= (u32)tmp << 16;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_FRN4);
	divn2_fracnum |= (u32)(tmp & 0x7f) << 24;
	if ((tmp & BIT(7)) != 0) {
            divn2_fracnum = -divn2_fracnum;
	}

	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_FRD1);
	divn2_fracden = tmp;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_FRD2);
	divn2_fracden |= (u32)tmp << 8;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_FRD3);
	divn2_fracden |= (u32)tmp << 16;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVN2_FRD4);
	divn2_fracden |= (u32)(tmp) << 24;


	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVNFRAC2);
	divn = (u32)tmp << 8;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVNFRAC3);
	divn |= (u32)tmp << 16;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVNFRAC4);
	divn |= (u32)(tmp & 0x7f) << 24;
	if ((tmp & BIT(7)) != 0) {
            divn = -divn;
	}
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_DIVNINT);
	divn |= (u64)(tmp & 0x7f) << 32;
	tmp = sit_95147_read_reg(pll, SIT_95147_PAGE_PLL_MISCXO);
	divn |= (u64)(tmp & BIT(4)) << (32+7-4);


	pll->divn = divn;
	pll->divn2_int = divn2_int;
	pll->divn2_fracnum = divn2_fracnum;
	pll->divn2_fracden = divn2_fracden;

	{
		// fetch one time the status to ignore inital lose of lock
		struct si_time_pll_internal_status status;
		struct si_time_pll_internal_input_status status_input;
		si_time_pll_get_status_internal(pll, pll->clock_input_pll_page, &status);
		si_time_pll_input_get_status_internal(pll, pll->clock_input, &status_input);
	}

	INIT_WORK(&pll->adjfreq_work, si_time_pll_clock_adjfine_work);

	INIT_WORK(&pll->watchdog_work, sit_95147_poll_status);

	timer_setup(&pll->watchdog, sit_95147_poll_status_timer, 0);
	pll->watchdog.expires = jiffies + POLL_TIME;
	add_timer(&pll->watchdog);

	if(device_create_file(dev, &dev_attr_clock_input)) {
		dev_err(dev, "could not crate sys attr\n");
		sit_95147_exit(pll);
		return -EFAULT;
	}

	dev_info(dev, "driver loaded");
	return 0;
}

void sit_95147_exit(struct si_time_95147 *pll) {

	if (del_timer_sync(&pll->watchdog))
		dev_dbg(pll->dev, "Had to wait for poll watchdog.\n");


}

#ifdef CONFIG_SPI

static int sit_95147_spi_probe(struct spi_device *spi) {
	struct si_time_95147 *ctx;
	struct si_time_95147_spi *ctx_spi;

	ctx = devm_kzalloc(&spi->dev, sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;
	ctx_spi = &ctx->bus.spi;
	spi_set_drvdata(spi, ctx);
	ctx_spi->spi = spi;
	ctx->dev = &spi->dev;
	ctx->bus_type = PLL_BUS_TYPE_SPI;


	return sit_95147_probe(ctx);
}

SPI_DRIVER_REMOVE_RETURN_TYPE sit_95147_spi_remove(struct spi_device *spi) {
	struct si_time_95147 *pll = spi_get_drvdata(spi);
	sit_95147_exit(pll);

	return SPI_DRIVER_REMOVE_RETURN_VALUE;
}

static const struct of_device_id sit95147_spi_of_match[] = {
	{
		.compatible	= "sit95147",
	},
	{ }
};
MODULE_DEVICE_TABLE(of, sit95147_spi_of_match);

static const struct spi_device_id sit95147_spi_id_table[] = {
	{
		.name		= "sit95147",
	},
	{ }
};
MODULE_DEVICE_TABLE(spi, sit95147_spi_id_table);

static struct spi_driver sit95147_spi_driver = {
	.driver = {
		.name = "sit95147",
		.of_match_table = sit95147_spi_of_match,
	},
	.id_table = sit95147_spi_id_table,
	.probe = sit_95147_spi_probe,
	.remove = sit_95147_spi_remove,
};

struct si_time_95147 *dev_get_si_time_pll_spi(struct device *dev) {
	if(IS_ERR_OR_NULL(dev)) {
		return NULL;
	}
	if(dev->driver == &sit95147_spi_driver.driver) {
		struct spi_device *spi = container_of(dev, struct spi_device, dev);
		struct si_time_95147 *pll = spi_get_drvdata(spi);
		if(IS_ERR_OR_NULL(pll)) {
			return NULL;
		}

		return pll;
	} else {
		return NULL;
	}
}
EXPORT_SYMBOL_GPL(dev_get_si_time_pll_spi);

#endif // end CONFIG_SPI

#ifdef CONFIG_I2C

int sit_95147_i2c_probe(struct i2c_client *client I2C_DRIVER_PROBE_OTHER_PARAMS){
	struct si_time_95147 *ctx;
	struct si_time_95147_i2c *ctx_i2c;

	ctx = devm_kzalloc(&client->dev, sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;
	ctx_i2c = &ctx->bus.i2c;

	i2c_set_clientdata(client, ctx);
	ctx_i2c->i2c = client;
	ctx->dev = &client->dev;
	ctx->bus_type = PLL_BUS_TYPE_I2C;

	return sit_95147_probe(ctx);
}
I2C_DRIVER_REMOVE_RETURN_TYPE sit_95147_i2c_remove(struct i2c_client *client){
	return I2C_DRIVER_REMOVE_RETURN_VALUE;
}


static const struct i2c_device_id sit95147_i2c_id_table[] = {
      { "sit95147" },
      { }
};

MODULE_DEVICE_TABLE(i2c, sit95147_i2c_id_table);

static struct i2c_driver sit95147_i2c_driver = {
      .driver = {
              .name   = "sit9514",
      },

      .id_table       = sit95147_i2c_id_table,
      .probe          = sit_95147_i2c_probe,
      .remove         = sit_95147_i2c_remove,
};
#endif // end CONFIG_I2C


static void sit95147_exit(void) {
#ifdef CONFIG_I2C
	i2c_del_driver(&sit95147_i2c_driver);
#endif

#ifdef CONFIG_SPI
	spi_unregister_driver(&sit95147_spi_driver);
#endif
}
module_exit(sit95147_exit);

static int sit95147_init(void) {
	int ret = 0;
#ifdef CONFIG_I2C
	ret = i2c_add_driver(&sit95147_i2c_driver);
	if(ret) {
		sit95147_exit();
		return ret;
	}
#endif

#ifdef CONFIG_SPI
	ret = spi_register_driver(&sit95147_spi_driver);
	if(ret) {
		sit95147_exit();
		return ret;
	}
#endif
	return 0;
}
module_init(sit95147_init);





static int sit_95147_find_device_rec(struct device *dev, void*pll) {
	if(dev->driver == &sit95147_spi_driver.driver) {
		get_device(dev);
		*(struct si_time_95147 **)pll = dev_get_drvdata(dev);
		return 1;
	}

	return device_for_each_child(dev, pll, sit_95147_find_device_rec);
}

struct si_time_95147 *sit_95147_get_device(struct device *dev) {
	struct si_time_95147 *pll = NULL;
	(void)sit_95147_find_device_rec(dev, &pll);
	return pll;
}
EXPORT_SYMBOL_GPL(sit_95147_get_device);


int si_time_pll_get_status(struct si_time_95147 *pll, struct si_time_pll_status *status){
	int ret = -1;

	//TODO


	return ret;
}
EXPORT_SYMBOL_GPL(si_time_pll_get_status);


static int si_time_pll_clock_adjfine_internal(struct si_time_95147 *pll, long scaled_ppm) {
	struct si_time_pll_internal_status status;
	s32 delta_scaled_ppm = scaled_ppm - pll->current_adjfreq;
	u32 delta_scaled_abs_ppm = abs(delta_scaled_ppm);
	const u64 PPM = 1000000u << 16;
	u64 num = mul_u64_u64_div_u64((u64)pll->divn2_int * pll->divn2_fracden + pll->divn2_fracnum, delta_scaled_abs_ppm, 1000000ULL << 16);
	u64 den = pll->divn2_fracden;
	u8 tmp_dco_function;
	u64 dco_code_int = num;
	u32 dco_code_frac = do_div(dco_code_int , den);

	si_time_pll_get_status_internal(pll, pll->clock_ajust_page, &status);
	pll->current_adjfreq = scaled_ppm;

	sit_95147_set_page(pll, pll->clock_ajust_page);


	if(dco_code_frac > (den >> 1)) {
		dco_code_int = dco_code_int + 1;
		dco_code_frac = den/2 - dco_code_frac;
	}
	if(status.holdover_mode) {
		dco_code_frac = mul_u64_u64_div_u64((u64)pll->divn, delta_scaled_abs_ppm, PPM);
	}

	sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_FRAC1, dco_code_frac & 0xff);
	sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_FRAC2, (dco_code_frac >> 8) & 0xff);
	sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_FRAC3, (dco_code_frac >> 16) & 0xff);
	sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_FRAC4, (dco_code_frac >> 24) & 0xff);

	if(!status.holdover_mode) {
		u8 bump3 = ((dco_code_int >> 8) & 0x3) << 6;
		bump3 |= 1u<<5 ; // enable dco sync mode
		sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_BUMP2, dco_code_int & 0xff);
		sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_BUMP3, bump3); // write the last two bits from the 10bit in value to the last two bits of the byte
	}

	tmp_dco_function = status.holdover_mode ? 1u<<1 : 0;
	sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_FUNCTION, tmp_dco_function);

	if(delta_scaled_ppm > 0) {
		sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_FUNCTION, tmp_dco_function | 0x08);
	} else {
		sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_FUNCTION, tmp_dco_function | 0x04);
	}
	mdelay(10);
	sit_95147_write_reg(pll, SIT_95147_PAGE_PLL_DCO_FUNCTION, tmp_dco_function);
	return 0;
}


static void si_time_pll_clock_adjfine_work(struct work_struct *work)
{
	struct si_time_95147* pll = container_of(work, struct si_time_95147, adjfreq_work);
	int ret;
	mutex_lock(&pll->lock);
	ret = si_time_pll_clock_adjfine_internal(pll, pll->pending_adjfreq);
	mutex_unlock(&pll->lock);

	(void)ret;
}


int si_time_pll_clock_adjfine(struct si_time_95147 *pll, long scaled_ppm) {
	pll->pending_adjfreq = scaled_ppm;

	schedule_work(&pll->adjfreq_work);

	return 0;
}
EXPORT_SYMBOL_GPL(si_time_pll_clock_adjfine);

MODULE_DESCRIPTION("SI Time PLL 95147 driver");
MODULE_AUTHOR("Norbert Summer <summer@oregano.at>");
MODULE_LICENSE("GPL v2");
MODULE_VERSION(DRV_VERSION);
