/**
 * Implements an SFF 8472 module
 * Doc: https://members.snia.org/document/dl/25916
*/

#include "oregano_tse_mac_sfp.h"
#include "swphy/swphy.h"

#include <linux/version.h>

#if (KERNEL_VERSION(6, 1, 0) <= LINUX_VERSION_CODE)
#include "compat/mdio-i2c.h"
#elif (KERNEL_VERSION(4, 14, 0) <= LINUX_VERSION_CODE)
#include "compat/mdio-i2c.h"
//TODO try to use upstream
//#include <linux/mdio/mdio-i2c.h>
#else
#include "compat/mdio-i2c.h"
#endif
#include "compat/compat.h"

#include <linux/netdevice.h>
#include <linux/phy_fixed.h>


struct oregano_phy {
	struct device *dev;
	struct oregano_mac_sfp *sfp;
	int addr;
	struct phy_device *phydev;
	struct fixed_phy_status status;
};

bool oregano_mac_sfp_connected(struct oregano_mac_sfp *sfp) {
	return sfp->state == SFP_PROBED;
}
const struct sfp_eeprom_base *oregano_mac_sfp_get_eeprom_base(struct oregano_mac_sfp *sfp) {
	return (struct sfp_eeprom_base*)sfp->eeprom.base;
}
const u8 *oregano_mac_sfp_get_eeprom_raw(struct oregano_mac_sfp *sfp) {
	return ( const u8 *)sfp->eeprom.base;
}

bool oregano_mac_sfp_is_sff8472_compliance(struct oregano_mac_sfp *sfp) {
	const u8 *eeprom = oregano_mac_sfp_get_eeprom_raw(sfp);
	return eeprom[SFP_SFF8472_COMPLIANCE] != 0;
}



static int fixed_mdio_read(struct mii_bus *bus, int phy_addr, int reg_num) {
	struct oregano_phy* oregano_phy = bus->priv;

	return swphy_read_reg(reg_num, &oregano_phy->status);
}

static int fixed_mdio_write(struct mii_bus *bus, int phy_addr, int reg_num,
			    u16 val)
{
	return 0;
}

static int fixed_mdio_bus_init(struct oregano_mac_sfp* sfp, struct oregano_phy *oregano_phy)
{
	int ret;

	sfp->mii_bus = mdiobus_alloc();
	if (sfp->mii_bus == NULL) {
		return -ENOMEM;
	}

	snprintf(sfp->mii_bus->id, MII_BUS_ID_SIZE, "fixed:%s-0", sfp->net_dev->name);
	sfp->mii_bus->name = "Fixed MDIO Bus";
	sfp->mii_bus->priv = oregano_phy;
	sfp->mii_bus->parent = &sfp->net_dev->dev;
	sfp->mii_bus->read = &fixed_mdio_read;
	sfp->mii_bus->write = &fixed_mdio_write;
	sfp->mii_bus->phy_mask = ~0;

	ret = mdiobus_register(sfp->mii_bus);
	if (ret)	{
		mdiobus_free(sfp->mii_bus);
		return ret;
	}

	return 0;
}

static struct phy_driver oregano_sfp_phy_driver = {
	.name = "oregano_sfp_fiber",
};


static struct phy_device *create_onboard_phy(struct oregano_mac_sfp *sfp) {
	struct net_device *ndev = sfp->net_dev;
	struct device *dev = &ndev->dev;
	int ret = 0;
	struct oregano_phy *oregano_phy = kzalloc(sizeof(struct oregano_phy), GFP_KERNEL);
	struct phy_device *phy_device;

	dev_dbg(dev, "create onboard_phy");

	if (!oregano_phy)
		return ERR_PTR(-ENOMEM);

	oregano_phy->dev = dev;
	oregano_phy->addr = 0;
	oregano_phy->status.link = 0;
	oregano_phy->status.speed = SPEED_1000;
	oregano_phy->status.duplex = DUPLEX_FULL;

	ret = fixed_mdio_bus_init(sfp, oregano_phy);
	if(ret) {
		dev_err(dev, "could not allocate onboard phy\n");
		kfree(oregano_phy);
		return ERR_PTR(ret);
	}
	phy_device = get_phy_device(sfp->mii_bus, oregano_phy->addr, false);
	if (IS_ERR(phy_device)) {
		dev_err(dev, "could not get_phy_device  errno: %ld\n", PTR_ERR(phy_device));
		mdiobus_unregister(sfp->mii_bus);
		mdiobus_free(sfp->mii_bus);
		kfree(oregano_phy);
		return ERR_PTR(-EINVAL);
	}
	phy_device->drv = &oregano_sfp_phy_driver;
	phy_device->priv = oregano_phy;

	/* propagate the fixed link values to struct phy_device */
	phy_device->link = oregano_phy->status.link;
	phy_device->speed = oregano_phy->status.speed;
	phy_device->duplex = oregano_phy->status.duplex;
	phy_device->pause = oregano_phy->status.pause;
	phy_device->asym_pause = oregano_phy->status.asym_pause;
	linkmode_set_bit(ETHTOOL_LINK_MODE_1000baseX_Full_BIT,
				phy_device->supported);

	oregano_phy->phydev = phy_device;

	phy_advertise_supported(phy_device);
	ret = phy_device_register(phy_device);
	if (ret) {
		dev_err(dev, "could not register phy device\n");
		phy_device_free(phy_device);
		mdiobus_unregister(sfp->mii_bus);
		mdiobus_free(sfp->mii_bus);
		sfp->mii_bus=NULL;
		kfree(oregano_phy);
		return ERR_PTR(ret);
	}

	return phy_device;
}



static int sfp_i2c_read(struct i2c_adapter *i2c, bool a2, u8 dev_addr, void *buf,
			size_t len);


static int sfp_i2c_read(struct i2c_adapter *i2c, bool a2, u8 dev_addr, void *buf,
			size_t len)
{
	struct i2c_msg msgs[2];
	u8 bus_addr = a2 ? 0x51 : 0x50;
	size_t block_size = 16;
	size_t this_len;
	int ret;

	msgs[0].addr = bus_addr;
	msgs[0].flags = 0;
	msgs[0].len = 1;
	msgs[0].buf = &dev_addr;
	msgs[1].addr = bus_addr;
	msgs[1].flags = I2C_M_RD;
	msgs[1].len = len;
	msgs[1].buf = buf;

	while (len) {
		this_len = len;
		if (this_len > block_size)
			this_len = block_size;

		msgs[1].len = this_len;

		ret = i2c_transfer(i2c, msgs, ARRAY_SIZE(msgs));
		if (ret < 0)
			return ret;

		if (ret != ARRAY_SIZE(msgs))
			break;

		msgs[1].buf += this_len;
		dev_addr += this_len;
		len -= this_len;
	}

	return msgs[1].buf - (u8 *)buf;
}

#if 0
//currently not needed function
static int sfp_i2c_write(struct i2c_adapter *i2c, bool a2, u8 dev_addr, void *buf,
	size_t len)
{
	struct i2c_msg msgs[1];
	u8 bus_addr = a2 ? 0x51 : 0x50;
	int ret;

	msgs[0].addr = bus_addr;
	msgs[0].flags = 0;
	msgs[0].len = 1 + len;
	msgs[0].buf = kmalloc(1 + len, GFP_KERNEL);
	if (!msgs[0].buf)
		return -ENOMEM;

	msgs[0].buf[0] = dev_addr;
	memcpy(&msgs[0].buf[1], buf, len);

	ret = i2c_transfer(i2c, msgs, ARRAY_SIZE(msgs));

	kfree(msgs[0].buf);

	if (ret < 0)
		return ret;

	return ret == ARRAY_SIZE(msgs) ? len : 0;
}
#endif

static int oregano_mac_sfp_read_eeprom(struct oregano_mac_sfp *sfp) {
	u8 *eeprom_raw = (u8 *)&sfp->eeprom;
	return sfp_i2c_read(sfp->i2c, false, 0, eeprom_raw, sizeof(sfp->eeprom));
}

void oregano_mac_sfp_exit(struct oregano_mac_sfp *sfp) {
	if(IS_ERR_OR_NULL(sfp->net_dev)) {
		sfp->state = SFP_NOT_INIT;
		return;
	}
	oregano_mac_sfp_exit_phy(sfp);

	sfp->state = SFP_NOT_INIT;
}

static bool oregano_phy_sfp_verify_eeprom(u8 cc_in, u8 *buf, unsigned int len)
{
	u8 cc;

	for (cc = 0; len; buf++, len--)
		cc += *buf;

	return cc == cc_in;
}

int oregano_mac_sfp_probe(struct device *dev, struct oregano_mac_sfp *sfp, struct net_device *net_dev, struct i2c_adapter *i2c_adapter, phy_interface_t interface) {
	//i2c_get_adapter(i2c_adapter->nr);
	sfp->i2c = i2c_adapter;
	sfp->net_dev = net_dev;
	sfp->interface = interface;
	return 0;
}

void oregano_mac_sfp_exit_phy(struct oregano_mac_sfp *sfp) {
	struct net_device *ndev = sfp->net_dev;
	struct phy_device *phydev = ndev->phydev;
	if(!IS_ERR_OR_NULL(phydev)) {
		phy_disconnect(phydev);
		phy_device_remove(phydev);
		phy_device_free(phydev);
		ndev->phydev=NULL;
	}
	if (!IS_ERR_OR_NULL(sfp->mii_bus)) {
		mdiobus_unregister(sfp->mii_bus);
		mdiobus_free(sfp->mii_bus);
		sfp->mii_bus = NULL;
	}
	sfp->state = SFP_NOT_INIT;
}

int oregano_mac_sfp_probe_phy(struct oregano_mac_sfp *sfp, void (*phy_handler)(struct net_device *ndev) ) {
	int phy_id = 22;
	struct net_device *ndev = sfp->net_dev;
	struct device *dev = &ndev->dev;
	struct phy_device *phydev = ndev->phydev;
	struct sfp_eeprom_base *eeprom_base = NULL;
	int ret = 0;

	dev_dbg(dev, "probe phy");

	if(sfp->state == SFP_PROBED && phydev) {
		pr_err("should not happen!!\n");
		oregano_mac_sfp_exit_phy(sfp);
	}

	ret = oregano_mac_sfp_read_eeprom(sfp);
	if(ret < 0) {
		pr_err("error reading sfp-eeprom\n");
		return ret;
	}

	eeprom_base = (struct sfp_eeprom_base *)&sfp->eeprom.base;
	if(!oregano_phy_sfp_verify_eeprom(eeprom_base->cc_base, sfp->eeprom.base, sizeof(sfp->eeprom.base)-1)) {
		pr_err("error invalid sfp-eeprom checksum\n");
		return -EFAULT;
	}

	if(eeprom_base->e1000_base_t) {
		sfp->mii_bus = mdio_i2c_alloc(dev, sfp->i2c);
		if(IS_ERR_OR_NULL(sfp->mii_bus)) {
			pr_err("can't alloc i2c mdio\n");
			sfp->mii_bus = NULL;
			return -ENOMEM;
		}

		sfp->mii_bus->name = "SFP I2C Bus";
		sfp->mii_bus->phy_mask = ~0;
		ret = mdiobus_register(sfp->mii_bus);
		if (ret < 0) {
			pr_err("can't register i2c mdio\n");
			mdiobus_free(sfp->mii_bus);
			sfp->mii_bus = NULL;
			return ret;
		}
		phydev = mdiobus_scan_c22(sfp->mii_bus, phy_id);
		if (IS_ERR_OR_NULL(phydev)) {
			pr_err("error no mdio phy found for base_t\n");
			oregano_mac_sfp_exit_phy(sfp);
			return -ENODEV;
		}
	} else if(eeprom_base->e1000_base_sx || eeprom_base->e1000_base_lx || true) {
		phydev = create_onboard_phy(sfp);
		if (IS_ERR_OR_NULL(phydev)) {
			pr_err("error could not init base_x phy\n");
			return -ENODEV;
		}
		phy_handler(ndev);
	} else if(eeprom_base->e10g_base_sr || eeprom_base->e10g_base_lr) {
		phydev = create_onboard_phy(sfp);
		if (IS_ERR_OR_NULL(phydev)) {
			pr_err("error could not init base_x phy\n");
			return -ENODEV;
		}
		phy_handler(ndev);
	} else {
		pr_err("not supported sfp module type");
		return -ENODEV;
	}
	phy_attached_info(phydev);
	phy_connect_direct(ndev, phydev, phy_handler, sfp->interface );

	ndev->phydev = phydev;
	sfp->state = SFP_PROBED;

	phy_start(phydev);

	return 0;
}

/**
 * sets the phy link status from the oregano mac link status register
*/
void oregano_mac_sfp_set_phylink(struct oregano_mac_sfp *sfp, bool link_status) {
	struct net_device *ndev = sfp->net_dev;
	struct phy_device *phydev = ndev->phydev;

	sfp->link_status = link_status;

	if(!IS_ERR_OR_NULL(phydev) /* && phydev->drv == &oregano_sfp_phy_driver*/ ) {
		struct sfp_eeprom_base *eeprom_base = (struct sfp_eeprom_base *)&sfp->eeprom.base;
		if(!eeprom_base->e1000_base_t) {
			struct oregano_phy *oregano_phy = phydev->priv;
			oregano_phy->status.link = link_status;
		}
	}
}
